from io import BytesIO
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement
from xml.dom import minidom


"""
xmlio can do input and output of XML, allowing it to be edited using ElementTree library
"""


# namespaces for when reparsing XML strings
REPARSING_NAMESPACES = (
    'xmlns:ali="http://www.niso.org/schemas/ali/1.0/" '
    + 'xmlns:jats="http://www.ncbi.nlm.nih.gov/JATS1" '
    + 'xmlns:mml="http://www.w3.org/1998/Math/MathML" '
    + 'xmlns:xlink="http://www.w3.org/1999/xlink"'
)


class CustomTreeBuilder(ElementTree.TreeBuilder):
    doctype_dict = {}

    def doctype(self, name, pubid, system):
        self.doctype_dict["name"] = name
        self.doctype_dict["pubid"] = pubid
        self.doctype_dict["system"] = system


def register_xmlns():
    """
    Register namespaces globally
    """
    ElementTree.register_namespace("ali", "http://www.niso.org/schemas/ali/1.0/")
    ElementTree.register_namespace("mml", "http://www.w3.org/1998/Math/MathML")
    ElementTree.register_namespace("xlink", "http://www.w3.org/1999/xlink")


def parse(filename, return_doctype_dict=False, return_processing_instructions=False):
    """
    to extract the doctype details from the file when parsed and return the data
    for later use, set return_doctype_dict to True
    """
    doctype_dict = {}
    processing_instructions = []

    # support BytesIO or actual files
    try:
        xml_bytes = filename.getvalue()
    except AttributeError:
        with open(filename, "rb") as open_file:
            xml_bytes = open_file.read()

    # collect processing instruction nodes using minidom
    with minidom.parseString(xml_bytes) as dom:
        for node in dom.childNodes:
            if isinstance(node, minidom.ProcessingInstruction):
                processing_instructions.append(node)

    # Assume greater than Python 3.2, get the doctype from the TreeBuilder
    tree_builder = CustomTreeBuilder()
    parser = ElementTree.XMLParser(target=tree_builder, encoding="utf-8")
    new_file = BytesIO(xml_bytes)
    tree = ElementTree.parse(new_file, parser)

    root = tree.getroot()

    doctype_dict = tree_builder.doctype_dict

    if return_doctype_dict is True and return_processing_instructions is True:
        return root, doctype_dict, processing_instructions
    if return_doctype_dict is True:
        return root, doctype_dict
    return root


def add_tag_before(tag_name, tag_text, parent_tag, before_tag_name):
    """
    Helper function to refactor the adding of new tags
    especially for when converting text to role tags
    """
    new_tag = Element(tag_name)
    new_tag.text = tag_text
    if get_first_element_index(parent_tag, before_tag_name):
        parent_tag.insert(
            get_first_element_index(parent_tag, before_tag_name) - 1, new_tag
        )
    return parent_tag


def get_first_element_index(root, tag_name):
    """
    In order to use Element.insert() in a convenient way,
    this function will find the first child tag with tag_name
    and return its index position
    The index can then be used to insert an element before or after the
    found tag using Element.insert()
    """
    tag_index = 1
    for tag in root:
        if tag.tag == tag_name:
            # Return the first one found if there is a match
            return tag_index
        tag_index = tag_index + 1
    # Default
    return None


def convert_xlink_href(root, name_map):

    xpath_list = [
        ".//graphic",
        ".//media",
        ".//inline-graphic",
        ".//self-uri",
        ".//ext-link",
    ]
    count = 0
    for xpath in xpath_list:
        for tag in root.findall(xpath):

            if tag.get("{http://www.w3.org/1999/xlink}href"):

                for key, value in name_map.items():
                    # Try to match the exact name first, and if not then
                    #  try to match it without the file extension
                    if tag.get("{http://www.w3.org/1999/xlink}href") == key:
                        tag.set("{http://www.w3.org/1999/xlink}href", value)
                        count += 1
                    elif (
                        tag.get("{http://www.w3.org/1999/xlink}href")
                        == key.split(".")[0]
                    ):
                        tag.set(
                            "{http://www.w3.org/1999/xlink}href", value.split(".")[0]
                        )
                        count += 1
    return count


def rewrite_subject_group(root, subjects, subject_group_type, overwrite=True):
    "add or rewrite subject tags inside subj-group tags"
    parent_tag_name = "subj-group"
    tag_name = "subject"
    wrap_tag_name = "article-categories"
    tag_attribute = "subj-group-type"
    # the parent tag where it should be found
    xpath_parent = ".//front/article-meta/article-categories"
    # the wraping tag in case article-categories does not exist
    xpath_article_meta = ".//front/article-meta"
    # the xpath to find the subject tags we are interested in
    xpath = './/{parent_tag_name}[@{tag_attribute}="{group_type}"]'.format(
        parent_tag_name=parent_tag_name,
        tag_attribute=tag_attribute,
        group_type=subject_group_type,
    )

    count = 0
    # get the parent tag
    parent_tag = root.find(xpath_parent)
    if parent_tag is None:
        # parent tag not found, add one
        wrap_tag = root.find(xpath_article_meta)
        article_categories_tag = SubElement(wrap_tag, wrap_tag_name)
        parent_tag = article_categories_tag
    insert_index = 0
    # iterate all tags to find the index of the first tag we are interested in
    if parent_tag is not None:
        for tag_index, tag in enumerate(parent_tag.findall("*")):
            if (
                tag.tag == parent_tag_name
                and tag.get(tag_attribute) == subject_group_type
            ):
                insert_index = tag_index
                if overwrite is True:
                    # if overwriting use the first one found
                    break
            # if not overwriting, use the last one found + 1
            if overwrite is not True:
                insert_index += 1
    # remove the tag if overwriting the existing values
    if overwrite is True:
        # remove all the tags
        for tag in root.findall(xpath):
            parent_tag.remove(tag)
    # add the subjects
    for subject in subjects:
        subj_group_tag = Element(parent_tag_name)
        subj_group_tag.set(tag_attribute, subject_group_type)
        subject_tag = SubElement(subj_group_tag, tag_name)
        subject_tag.text = subject
        parent_tag.insert(insert_index, subj_group_tag)
        count += 1
        insert_index += 1
    return count


def output(root, output_type="JATS", doctype_dict=None, processing_instructions=None):

    if doctype_dict is not None:
        publicId = doctype_dict.get("pubid")
        systemId = doctype_dict.get("system")
        qualifiedName = doctype_dict.get("name")
    elif output_type == "JATS":
        publicId = (
            "-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange "
            "DTD v1.1d3 20150301//EN"
        )
        systemId = "JATS-archivearticle1.dtd"
        qualifiedName = "article"
    else:
        publicId = None
        systemId = None
        qualifiedName = "article"

    encoding = "UTF-8"

    doctype = build_doctype(qualifiedName, publicId, systemId)

    return output_root(root, doctype, encoding, processing_instructions)


def output_root(root, doctype, encoding, processing_instructions=None):
    rough_string = ElementTree.tostring(root, encoding)

    reparsed = minidom.parseString(rough_string)
    if doctype:
        reparsed.insertBefore(doctype, reparsed.documentElement)

    if processing_instructions:
        for pi_node in processing_instructions:
            reparsed.insertBefore(pi_node, reparsed.documentElement)

    # reparsed_string =  reparsed.toprettyxml(indent="\t", encoding = encoding)
    reparsed_string = reparsed.toxml(encoding=encoding)

    return reparsed_string


def build_doctype(qualifiedName, publicId=None, systemId=None, internalSubset=None):
    """
    Instantiate an ElifeDocumentType, a subclass of minidom.DocumentType, with
    some properties so it is more testable
    """
    doctype = ElifeDocumentType(qualifiedName)
    doctype._identified_mixin_init(publicId, systemId)
    if internalSubset:
        doctype.internalSubset = internalSubset
    return doctype


class ElifeDocumentType(minidom.DocumentType):
    """
    Override minidom.DocumentType in order to get
    double quotes in the DOCTYPE rather than single quotes
    """

    def writexml(self, writer, indent="", addindent="", newl=""):
        assert self.name is not None, "the `minidom.DocumentType.name` property has not been set. see `output` and `build_doctype`."

        writer.write("<!DOCTYPE ")

        # Throws TypeError if self.name is None. only happens in Python 3.
        writer.write(self.name)

        if self.publicId:
            writer.write(
                '%s PUBLIC "%s"%s  "%s"' % (newl, self.publicId, newl, self.systemId)
            )
        elif self.systemId:
            writer.write('%s SYSTEM "%s"' % (newl, self.systemId))

        if self.internalSubset is not None:
            writer.write(" [")
            writer.write(self.internalSubset)
            writer.write("]")
        writer.write(">" + newl)


def append_minidom_xml_to_elementtree_xml(
    parent, xml, recursive=False, attributes=None, child_attributes=False
):
    """
    Recursively,
    Given an ElementTree.Element as parent, and a minidom instance as xml,
    append the tags and content from xml to parent
    Used primarily for adding a snippet of XML with <italic> tags
    attributes: a list of attribute names to copy
    """

    # Get the root tag name
    if recursive is False:
        tag_name = xml.documentElement.tagName
        node = xml.getElementsByTagName(tag_name)[0]
        new_elem = SubElement(parent, tag_name)
        if attributes:
            for attribute in attributes:
                if xml.documentElement.hasAttribute(attribute):
                    new_elem.set(attribute, xml.documentElement.getAttribute(attribute))
    else:
        node = xml
        tag_name = node.tagName
        new_elem = parent
        # copy child tag attributes if present
        if child_attributes and node.hasAttributes():
            for name, value in node.attributes.items():
                new_elem.set(name, value)

    new_elem_sub = None
    i = 0
    for child_node in node.childNodes:
        if child_node.nodeName == "#text":
            if not new_elem.text and i <= 0:
                new_elem.text = child_node.nodeValue
            elif new_elem_sub is not None and (not new_elem.text and i > 0):
                new_elem_sub.tail = child_node.nodeValue
            elif new_elem_sub is not None:
                new_elem_sub.tail = child_node.nodeValue

        elif child_node.childNodes is not None:
            new_elem_sub = SubElement(new_elem, child_node.tagName)
            new_elem_sub = append_minidom_xml_to_elementtree_xml(
                new_elem_sub, child_node, True, attributes, child_attributes
            )

        i = i + 1

    # Debug
    # encoding = 'utf-8'
    # rough_string = ElementTree.tostring(parent, encoding)
    # print rough_string

    return parent


if __name__ == "__main__":

    # Sample usage
    article_xml_filenames = ["sample-xml/elife-kitchen-sink.xml"]

    for filename in article_xml_filenames:
        print("converting " + filename)

        register_xmlns()

        root = parse(filename)

        reparsed_string = output(root)

        print(reparsed_string)


def reparsed_tag(
    tag_name, tag_string, namespaces=REPARSING_NAMESPACES, attributes_text=""
):
    """given tag content and attributes, reparse to a minidom tag"""
    open_tag_parts = [
        value for value in [tag_name, namespaces, attributes_text] if value
    ]
    open_tag = "<%s>" % " ".join(open_tag_parts)
    close_tag = "</%s>" % tag_name
    tagged_string = "%s%s%s" % (open_tag, tag_string, close_tag)
    return minidom.parseString(tagged_string.encode("utf-8"))



import time
import calendar
import re
from collections import OrderedDict
from slugify import slugify
from bs4 import Comment


def subject_slug(subject, stopwords=None):
    "create a slug for a subject value"
    if stopwords is None:
        stopwords = ["and", "of"]
    if not subject:
        return subject
    return slugify(subject, stopwords=stopwords)


def first(value):
    if value is None:
        return None
    "returns the first element of an iterable, swallowing index errors and returning None"
    try:
        return value[0]
    except IndexError:
        return None


def firstnn(value):
    "returns the first non-nil value within given iterable"
    return first(list(filter(None, value)))


def strip_strings(value):
    def strip_string(value):
        if hasattr(value, "strip"):
            return value.strip()
        return value

    if isinstance(value, list):
        return list(map(strip_string, value))
    return strip_string(value)


def strip_punctuation_space(value):
    "Strip excess whitespace prior to punctuation."

    def strip_punctuation(string):
        replacement_list = (
            (" .", "."),
            (" :", ":"),
            ("( ", "("),
            (" )", ")"),
        )
        for match, replacement in replacement_list:
            string = string.replace(match, replacement)
        return string

    if value is None:
        return None
    if isinstance(value, list):
        return [strip_punctuation(v) for v in value]
    return strip_punctuation(value)


def join_sentences(string1, string2, glue="."):
    "concatenate two sentences together with punctuation glue"
    if not string1 or string1 == "":
        return string2
    if not string2 or string2 == "":
        return string1
    # both are strings, continue joining them together with the glue and whitespace
    new_string = string1.rstrip()
    if not new_string.endswith(glue):
        new_string += glue
    new_string += " " + string2.lstrip()
    return new_string


def coerce_to_int(val, default=0xDEADBEEF):
    """
    Attempts to cast given value to an integer, return the original value if
    failed or the default if one provided.
    """
    try:
        return int(val)
    except (TypeError, ValueError):
        if default != 0xDEADBEEF:
            return default
        return val


def nullify(function):
    "Decorator. If empty list, returns None, else list."

    def wrapper(*args, **kwargs):
        value = function(*args, **kwargs)
        if isinstance(value, list) and len(value) == 0:
            return None
        return value

    return wrapper


def strippen(function):
    "Decorator. Strip excess whitespace from return value."

    def wrapper(*args, **kwargs):
        return strip_strings(function(*args, **kwargs))

    return wrapper


def inten(function):
    "Decorator. Attempts to convert return value to int"

    def wrapper(*args, **kwargs):
        return coerce_to_int(function(*args, **kwargs))

    return wrapper


def clean_whitespace(value):
    if not value:
        return value
    if hasattr(value, "lstrip"):
        value = value.lstrip("\n ")
    if hasattr(value, "rstrip"):
        value = value.rstrip("\n ")
    value = value.replace("\n", " ")
    return value


def date_struct(year, month, day, tz="UTC"):
    """
    Given year, month and day numeric values and a timezone
    convert to structured date object
    """
    ymdtz = (year, month, day, tz)
    if None in ymdtz:
        # logger.debug("a year, month, day or tz value was empty: %s" % str(ymdtz))
        return None  # return early if we have a bad value
    try:
        return time.strptime("%s-%s-%s %s" % ymdtz, "%Y-%m-%d %Z")
    except (TypeError, ValueError):
        # logger.debug("date failed to convert: %s" % str(ymdtz))
        return None


def date_struct_nn(year, month, day, tz="UTC"):
    """
    Assemble a date object but if day or month is none set them to 1
    to make it easier to deal with partial dates
    """
    if not day:
        day = 1
    if not month:
        month = 1
    return date_struct(year, month, day, tz)


def date_text(date_struct):
    # looks like: January 01, 2015
    return time.strftime("%B %d, %Y", date_struct) if date_struct else None


# TODO: if these are being converted to integers perhaps they shouldn't end with '_text' ?


@inten
def day_text(date_struct):
    return time.strftime("%d", date_struct) if date_struct else None


@inten
def month_text(date_struct):
    return time.strftime("%m", date_struct) if date_struct else None


@inten
def year_text(date_struct):
    return time.strftime("%Y", date_struct) if date_struct else None


def date_timestamp(date_struct):
    try:
        return calendar.timegm(date_struct)
    except (TypeError, ValueError):
        return None


def paragraphs(tags):
    "Given a list of tags, only return the paragraph tags"
    return list(filter(lambda tag: tag.name == "p", tags))


def convert_testing_doi(doi):
    if doi is None:
        return doi
    parts = doi.split(".")
    new_msid = parts[-1][-5:]
    return ".".join(parts[0:-1] + [new_msid])


def starts_with_doi(tag):
    return bool(node_text(tag).strip().startswith("DOI:"))


def paragraph_is_only_doi(tag):
    return bool(
        node_text(tag).strip().startswith("http://dx.doi.org")
        and " " not in node_text(tag).strip()
        and node_contents_str(tag).startswith('<ext-link ext-link-type="doi"')
    )


def doi_uri_to_doi(value):
    "Strip the uri schema from the start of DOI URL strings"
    if value is None:
        return value
    replace_values = [
        "http://dx.doi.org/",
        "https://dx.doi.org/",
        "http://doi.org/",
        "https://doi.org/",
    ]
    for replace_value in replace_values:
        value = value.replace(replace_value, "")
    return value


def doi_to_doi_uri(value):
    "Turn DOI into a valid uri"
    if value is None:
        return value
    value = "https://doi.org/" + value
    return value


def remove_doi_paragraph(tags):
    "Given a list of tags, only return those whose text doesn't start with 'DOI:'"
    p_tags = list(filter(lambda tag: not starts_with_doi(tag), tags))
    p_tags = list(filter(lambda tag: not paragraph_is_only_doi(tag), p_tags))
    return p_tags


def orcid_uri_to_orcid(value):
    "Strip the uri schema from the start of ORCID URL strings"
    if value is None:
        return value
    replace_values = ["http://orcid.org/", "https://orcid.org/"]
    for replace_value in replace_values:
        value = value.replace(replace_value, "")
    return value


def remove_tag_from_tag(tag, nodename):
    if not nodename:
        return tag
    unwanted_tags = extract_nodes(tag, nodename)
    for unwanted_tag in unwanted_tags:
        unwanted_tag.decompose()
    return tag


def component_acting_parent_tag(parent_tag, tag):
    """
    Only intended for use in getting components, look for tag name of fig-group
    and if so, find the first fig tag inside it as the acting parent tag
    """
    if parent_tag.name == "fig-group":
        if len(tag.find_previous_siblings("fig")) > 0:
            acting_parent_tag = first(extract_nodes(parent_tag, "fig"))
        else:
            # Do not return the first fig as parent of itself
            return None
    else:
        acting_parent_tag = parent_tag
    return acting_parent_tag


#
#
#


def extract_nodes(soup, nodename, attr=None, value=None):
    """
    Returns a list of tags (nodes) from the given soup matching the given nodename.
    If an optional attribute and value are given, these are used to filter the results
    further."""
    tags = soup.find_all(nodename)
    if attr is not None and value is not None:
        return list(filter(lambda tag: tag.get(attr) == value, tags))
    return list(tags)


def node_text(tag):
    "Returns the text contents of a tag"
    return getattr(tag, "text", None)


def node_contents_str(tag):
    """
    Return the contents of a tag, including it's children, as a string.
    Does not include the root/parent of the tag.
    """
    if not tag:
        return None
    tag_string = ""
    for child_tag in tag.children:
        if isinstance(child_tag, Comment):
            # BeautifulSoup does not preserve comment tags, add them back
            tag_string += "<!--%s-->" % str(child_tag)
        else:
            tag_string += str(child_tag)
    return tag_string if tag_string != "" else None


def first_parent(tag, nodename):
    """
    Given a beautiful soup tag, look at its parents and return the first
    tag name that matches nodename or the list nodename
    """
    if nodename is not None and isinstance(nodename, str):
        nodename = [nodename]
    if tag and tag.parents:
        return first(list(filter(lambda tag: tag.name in nodename, tag.parents)))
    return None


def tag_ordinal(tag):
    """
    Given a beautiful soup tag, look at the tags of the same name that come before it
    to get the tag ordinal. For example, if it is tag name fig
    and two fig tags are before it, then it is the third fig (3)
    """
    return len(tag.find_all_previous(tag.name)) + 1


def tag_fig_ordinal(tag):
    """
    Meant for finding the position of fig tags with respect to whether
    they are for a main figure or a child figure
    """
    if "specific-use" not in tag.attrs:
        # Look for tags with no "specific-use" attribute
        return (
            len(
                list(
                    filter(
                        lambda tag: "specific-use" not in tag.attrs,
                        tag.find_all_previous(tag.name),
                    )
                )
            )
            + 1
        )
    return None


def tag_sibling_ordinal(tag):
    """
    Given a beautiful soup tag, count the same tags in its siblings
    to get a sibling "local" ordinal value. This is useful in counting
    child figures within a fig-group, for example
    """
    return len(tag.find_previous_siblings(tag.name)) + 1


def tag_limit_sibling_ordinal(tag, stop_tag_name):
    """
    Count previous tags of the same name until it
    reaches a tag name of type stop_tag, then stop counting
    """
    tag_count = 1
    for prev_tag in tag.previous_elements:
        if prev_tag.name == tag.name:
            tag_count += 1
        if prev_tag.name == stop_tag_name:
            break

    return tag_count


def tag_subarticle_sibling_ordinal(tag):
    return tag_limit_sibling_ordinal(tag, "sub-article")


def tag_appendix_sibling_ordinal(tag):
    return tag_limit_sibling_ordinal(tag, "app")


def tag_media_sibling_ordinal(tag):
    """
    Count sibling ordinal differently depending on if the
    mimetype is video or not
    """
    if hasattr(tag, "name") and tag.name != "media":
        return None

    nodenames = ["fig", "supplementary-material", "sub-article"]
    first_parent_tag = first_parent(tag, nodenames)

    sibling_ordinal = None

    if first_parent_tag:
        # Start counting at 0
        sibling_ordinal = 0
        for media_tag in first_parent_tag.find_all(tag.name):
            if "mimetype" in tag.attrs and tag["mimetype"] == "video":
                # Count all video type media tags
                if "mimetype" in media_tag.attrs and tag["mimetype"] == "video":
                    sibling_ordinal += 1
                if media_tag == tag:
                    break

            else:
                # Count all non-video type media tags
                if ("mimetype" not in media_tag.attrs) or (
                    "mimetype" in media_tag.attrs and tag["mimetype"] != "video"
                ):
                    sibling_ordinal += 1
                if media_tag == tag:
                    break
    else:
        # Start counting at 1
        sibling_ordinal = 1
        for prev_tag in tag.find_all_previous(tag.name):
            if not first_parent(prev_tag, nodenames):
                if "mimetype" in tag.attrs and tag["mimetype"] == "video":
                    # Count all video type media tags
                    if (
                        supp_asset(prev_tag) == supp_asset(tag)
                        and "mimetype" in prev_tag.attrs
                    ):
                        sibling_ordinal += 1
                else:
                    if (
                        supp_asset(prev_tag) == supp_asset(tag)
                        and "mimetype" not in prev_tag.attrs
                    ):
                        sibling_ordinal += 1

    return sibling_ordinal


def tag_supplementary_material_sibling_ordinal(tag):
    """
    Strategy is to count the previous supplementary-material tags
    having the same asset value to get its sibling ordinal.
    The result is its position inside any parent tag that
    are the same asset type
    """
    if hasattr(tag, "name") and tag.name != "supplementary-material":
        return None

    nodenames = ["fig", "media", "sub-article"]
    first_parent_tag = first_parent(tag, nodenames)

    sibling_ordinal = 1

    if first_parent_tag:
        # Within the parent tag of interest, count the tags
        #  having the same asset value
        for supp_tag in first_parent_tag.find_all(tag.name):
            if tag == supp_tag:
                # Stop once we reach the same tag we are checking
                break
            if supp_asset(supp_tag) == supp_asset(tag):
                sibling_ordinal += 1

    else:
        # Look in all previous elements that do not have a parent
        #  and count the tags having the same asset value
        for prev_tag in tag.find_all_previous(tag.name):
            if not first_parent(prev_tag, nodenames):
                if supp_asset(prev_tag) == supp_asset(tag):
                    sibling_ordinal += 1

    return sibling_ordinal


def supp_asset(tag):
    """
    Given a supplementary-material tag, the asset value depends on
    its label text. This also informs in what order (its ordinal) it
    has depending on how many of each type is present
    """
    # Default
    asset = "supp"
    if first(extract_nodes(tag, "label")):
        label_text = node_text(first(extract_nodes(tag, "label"))).lower()
        # Keyword match the label
        if label_text.find("code") > 0:
            asset = "code"
        elif label_text.find("data") > 0:
            asset = "data"
    return asset


def copy_attribute(source, source_key, destination, destination_key=None):
    if destination_key is None:
        destination_key = source_key
    if source is not None:
        if source is not None and destination is not None and source_key in source:
            destination[destination_key] = source[source_key]


def first_node_str_contents(soup, nodename, attr=None, value=None):
    return node_contents_str(
        first(extract_nodes(soup, nodename, attr=attr, value=value))
    )


def set_if_value(dictionary, key, value):
    if value is not None:
        dictionary[key] = value


def prune_dict_of_none_values(dictionary):
    return dict((k, v) for k, v in dictionary.items() if v is not None)


def text_to_title(value):
    """when a title is required, generate one from the value"""
    title = None
    if not value:
        return title
    words = value.split(" ")
    keep_words = []
    for word in words:
        if word.endswith(".") or word.endswith(":"):
            keep_words.append(word)
            if len(word) > 1 and "<italic>" not in word and "<i>" not in word:
                break
        else:
            keep_words.append(word)
    if len(keep_words) > 0:
        title = " ".join(keep_words)
        if title.split(" ")[-1] != "spp.":
            title = title.rstrip(" .:")
    return title


def rstrip_punctuation(value):
    "strip punctuation from the end of a label or title"
    if not value:
        return value
    return value.rstrip(".:")


def escape_unmatched_angle_brackets(string, allowed_tag_fragments=()):
    """
    In order to make an XML string less malformed, escape
    unmatched less than tags that are not part of an allowed tag
    Note: Very, very basic, and do not try regex \1 style replacements
      on unicode ever again! Instead this uses string replace
    allowed_tag_fragments is a tuple of tag name matches for use with startswith()
    """
    if not string:
        return string

    # Split string on tags
    tags = re.split("(<.*?>)", string)
    # print tags

    for i, val in enumerate(tags):
        # Use angle bracket character counts to find unmatched tags
        #  as well as our allowed_tags list to ignore good tags

        if val.count("<") == val.count(">") and not val.startswith(
            allowed_tag_fragments
        ):
            val = val.replace("<", "&lt;")
            val = val.replace(">", "&gt;")
        else:
            # Count how many unmatched tags we have
            while val.count("<") != val.count(">"):
                if val.count("<") != val.count(">") and val.count("<") > 0:
                    val = val.replace("<", "&lt;", 1)
                elif val.count("<") != val.count(">") and val.count(">") > 0:
                    val = val.replace(">", "&gt;", 1)
            if val.count("<") == val.count(">") and not val.startswith(
                allowed_tag_fragments
            ):
                # Send it through again in case there are nested unmatched tags
                val = escape_unmatched_angle_brackets(val, allowed_tag_fragments)

        tags[i] = val

    return "".join(tags)


def escape_ampersand(string):
    """
    Quick convert unicode ampersand characters not associated with
    a numbered entity or not starting with allowed characters to a plain &amp;
    """
    if not string:
        return string
    start_with_match = r"(\#x(....);|\#\d+?;|lt;|gt;|amp;)"
    # The pattern below is match & that is not immediately followed by #
    string = re.sub(r"&(?!" + start_with_match + ")", "&amp;", string)
    return string


def remove_tag(tag_name, string):
    "remove open tag and its attributes and remove close tag from an XML string"
    pattern = r"\n*</?%s.*?>\n*" % tag_name
    return re.sub(pattern, "", string) if string else ""


def remove_tag_and_text(tag_name, string):
    "remove open tag and close tag and the contents between them from an XML string"
    pattern = r"\n*<%s.*?>.*?</%s>\n*" % (tag_name, tag_name)
    return remove_tag(tag_name, re.sub(pattern, "", string)) if string else ""


def article_author_person(ref_author):
    return generate_author_person(ref_author, "article_author")


def references_author_person(ref_author):
    return generate_author_person(ref_author, "reference_author")


def generate_author_person(ref_author, preferred_format):
    "Create a person data structure, its preferred value depends on the preferred_format type"
    author_json = OrderedDict()
    author_json["type"] = "person"
    author_name = OrderedDict()

    if preferred_format == "article_author":
        author_name["preferred"] = author_preferred_name(
            ref_author.get("surname"),
            ref_author.get("given-names"),
            ref_author.get("suffix"),
        )
    elif preferred_format == "reference_author":
        author_name["preferred"] = reference_author_preferred_name(
            ref_author.get("surname"),
            ref_author.get("given-names"),
            ref_author.get("suffix"),
        )

    author_name["index"] = author_index_name(
        ref_author.get("surname"),
        ref_author.get("given-names"),
        ref_author.get("suffix"),
    )
    author_json["name"] = author_name
    return author_json


def reference_author_preferred_name(surname, given_names, suffix):
    # join surname given_names suffix, in that order, with a space character
    return " ".join(
        [element for element in [surname, given_names, suffix] if element is not None]
    )


def author_preferred_name(surname, given_names, suffix):
    return " ".join(
        [element for element in [given_names, surname, suffix] if element is not None]
    )


def author_index_name(surname, given_names, suffix):
    index_name = ", ".join(
        [element for element in [surname, given_names, suffix] if element is not None]
    )
    return index_name


def list_type_prefix(list_type):
    """from a JATS list list-type attribute return a list style prefix"""
    if list_type:
        if list_type == "simple":
            return "none"
        if list_type == "order":
            return "number"
        return list_type
    return "none"



from elifetools.utils import first, firstnn, extract_nodes, node_contents_str

"""
rawParser.py extracts and returns the nodes from the article xml using BeautifulSoup so that functionality at higher levels may use and combine them as neccessary.

"""


def article_meta(soup):
    return first(extract_nodes(soup, "article-meta"))


def article_title(soup):
    return first(extract_nodes(soup, "article-title"))


def title(soup):
    return first(extract_nodes(soup, "title"))


def abstract(soup, abstract_type=None):
    if abstract_type:
        return extract_nodes(
            soup, "abstract", attr="abstract-type", value=abstract_type
        )
    else:
        return extract_nodes(soup, "abstract")


def article_id(soup, pub_id_type=None):
    if pub_id_type:
        return extract_nodes(soup, "article-id", attr="pub-id-type", value=pub_id_type)
    else:
        return extract_nodes(soup, "article-id")


def meta_article_id(soup, pub_id_type):
    tags = article_id(soup, pub_id_type)
    # the first article-id tag whose parent is article-meta
    return first([tag for tag in tags if tag.parent.name == "article-meta"])


def doi(soup):
    return meta_article_id(soup, pub_id_type="doi")


def publisher_id(soup):
    return meta_article_id(soup, pub_id_type="publisher-id")


def journal_id(soup):
    # the first non-nil tag
    return firstnn(
        extract_nodes(soup, "journal-id", attr="journal-id-type", value="publisher-id")
    )


def journal_title(soup):
    return first(extract_nodes(soup, "journal-title"))


def journal_issn(soup, pub_format, pub_type):
    if pub_format is None and pub_type is None:
        # return the first issn tag found regardless of which type
        return first(extract_nodes(soup, "issn"))
    elif pub_format is not None:
        return first(
            extract_nodes(soup, "issn", attr="publication-format", value=pub_format)
        )
    elif pub_type is not None:
        return first(extract_nodes(soup, "issn", attr="pub-type", value=pub_type))


def publisher(soup):
    return first(extract_nodes(soup, "publisher-name"))


def article_type(soup):
    # returns raw data, just that the data doesn't contain any BS nodes
    return first(extract_nodes(soup, "article")).get("article-type")


def pub_date(soup, date_type=None, pub_type=None):
    if date_type is not None:
        return extract_nodes(soup, "pub-date", attr="date-type", value=date_type)
    elif pub_type is not None:
        return extract_nodes(soup, "pub-date", attr="pub-type", value=pub_type)
    else:
        return extract_nodes(soup, "pub-date")


def date(soup, date_type=None):
    if date_type is not None:
        return extract_nodes(soup, "date", attr="date-type", value=date_type)
    else:
        return extract_nodes(soup, "date")


def history_date(soup, date_type):
    date_tags = date(soup, date_type)
    return first([tag for tag in date_tags if tag.parent.name == "history"])


def day(soup):
    return first(extract_nodes(soup, "day"))


def month(soup):
    return first(extract_nodes(soup, "month"))


def year(soup):
    return first(extract_nodes(soup, "year"))


def keyword_group(soup):
    return extract_nodes(soup, "kwd-group")


def acknowledgements(soup):
    return first(extract_nodes(soup, "ack"))


def conflict(soup):
    conflict_tags = extract_nodes(soup, "fn", attr="fn-type", value="conflict")
    conflict_tags += extract_nodes(soup, "fn", attr="fn-type", value="COI-statement")
    return conflict_tags


def permissions(soup):
    # a better selector might be "article-meta.permissions"
    return extract_nodes(soup, "permissions")


def article_permissions(soup):
    # a better selector might be "article-meta.permissions"
    permissions_tags = permissions(soup)
    return first([tag for tag in permissions_tags if tag.parent.name == "article-meta"])


def licence(soup):
    return extract_nodes(soup, "license")


def licence_p(soup):
    return extract_nodes(soup, "license-p")


def licence_url(soup):
    "License url attribute of the license tag"
    if licence(soup):
        return first(licence(soup)).get("xlink:href")


def attrib(soup):
    return extract_nodes(soup, "attrib")


def copyright_statement(soup):
    return first(extract_nodes(soup, "copyright-statement"))


def copyright_year(soup):
    return first(extract_nodes(soup, "copyright-year"))


def copyright_holder(soup):
    return first(extract_nodes(soup, "copyright-holder"))


def funding_statement(soup):
    return first(extract_nodes(soup, "funding-statement"))


def affiliation(soup):
    return extract_nodes(soup, "aff")


def research_organism_keywords(soup):
    tags = first(
        extract_nodes(
            soup, "kwd-group", attr="kwd-group-type", value="research-organism"
        )
    )
    if not tags:
        return None
    return [tag for tag in tags if tag.name == "kwd"] or None


def author_keywords(soup):
    # A few articles have kwd-group with no kwd-group-type, so account for those
    tags = extract_nodes(soup, "kwd-group")
    keyword_tags = []
    for tag in tags:
        if (
            tag.get("kwd-group-type") == "author-keywords"
            or tag.get("kwd-group-type") is None
        ):
            keyword_tags += [tag for tag in tag if tag.name == "kwd"]
    return keyword_tags


def subject_area(soup, subject_group_type=None):
    # Supports all subject areas or just particular ones filtered by
    subject_area_tags = []
    tags = extract_nodes(soup, "subject")

    subject_area_tags = [
        tag
        for tag in tags
        if tag.parent.name == "subj-group"
        and tag.parent.parent.name == "article-categories"
        and tag.parent.parent.parent.name == "article-meta"
    ]
    if subject_group_type:
        subject_area_tags = [
            tag
            for tag in tags
            if tag.parent.get("subj-group-type") == subject_group_type
        ]
    return subject_area_tags


def full_subject_area(soup, subject_group_type=None):

    subject_group_tags = extract_nodes(soup, "subj-group")
    subject_group_tags = [
        tag
        for tag in subject_group_tags
        if tag.parent.name == "article-categories"
        and tag.parent.parent.name == "article-meta"
    ]

    if subject_group_type:
        subject_group_tags = list(
            filter(lambda tag: tag.get("subj-group-type" == subject_group_type))
        )

    return subject_group_tags


def custom_meta(soup, meta_name=None):
    custom_meta_tags = extract_nodes(soup, "custom-meta")
    if meta_name is not None:
        custom_meta_tags = [
            tag
            for tag in custom_meta_tags
            if node_contents_str(first(extract_nodes(tag, "meta-name"))) == meta_name
        ]
    return custom_meta_tags


def display_channel(soup):
    return subject_area(soup, subject_group_type="display-channel")


def sub_display_channel(soup):
    return subject_area(soup, subject_group_type="sub-display-channel")


def category(soup):
    return subject_area(soup, subject_group_type="heading")


def related_article(soup):
    related_article_tags = extract_nodes(soup, "related-article")
    return [tag for tag in related_article_tags if tag.parent.name == "article-meta"]


def mixed_citations(soup):
    return extract_nodes(soup, "mixed-citation")


def related_object(soup):
    return extract_nodes(soup, "related-object")


def object_id(soup, pub_id_type):
    return extract_nodes(soup, "object-id", attr="pub-id-type", value=pub_id_type)


def pub_history(soup):
    return extract_nodes(soup, "pub-history")


def event(soup):
    return extract_nodes(soup, "event")


def event_desc(soup):
    return extract_nodes(soup, "event-desc")


def label(soup):
    return first(extract_nodes(soup, "label"))


def contributors(soup):
    return extract_nodes(soup, "contrib")


def article_contributors(soup):
    article_meta_tag = article_meta(soup)
    if article_meta_tag:
        contributor_tags = extract_nodes(article_meta_tag, ["contrib", "on-behalf-of"])
        return [tag for tag in contributor_tags if tag.parent.name == "contrib-group"]


def authors(soup, contrib_type="author"):
    if contrib_type:
        return extract_nodes(soup, "contrib", attr="contrib-type", value=contrib_type)
    else:
        return extract_nodes(soup, "contrib")


def caption(soup):
    return first(extract_nodes(soup, "caption"))


def author_notes(soup):
    return first(extract_nodes(soup, "author-notes"))


def corresp(soup):
    return extract_nodes(soup, "corresp")


def fn_group(soup, content_type=None):
    if content_type:
        return extract_nodes(soup, "fn-group", attr="content-type", value=content_type)
    else:
        return extract_nodes(soup, "fn-group")


def fn(soup):
    return extract_nodes(soup, "fn")


def media(soup):
    return extract_nodes(soup, "media")


def inline_graphic(soup):
    return extract_nodes(soup, "inline-graphic")


def graphic(soup):
    return extract_nodes(soup, "graphic")


def self_uri(soup):
    return extract_nodes(soup, "self-uri")


def supplementary_material(soup):
    return extract_nodes(soup, "supplementary-material")


#
# authors
#


def contrib_id(soup):
    return extract_nodes(soup, "contrib-id")


def email(soup):
    return extract_nodes(soup, "email")


def phone(soup):
    return extract_nodes(soup, "phone")


def bio(soup):
    return extract_nodes(soup, "bio")


#
# references
#


def ref_list(soup):
    return extract_nodes(soup, "ref")


def volume(soup):
    return extract_nodes(soup, "volume")


def issue(soup):
    return extract_nodes(soup, "issue")


def elocation_id(soup):
    return extract_nodes(soup, "elocation-id")


def fpage(soup):
    return extract_nodes(soup, "fpage")


def lpage(soup):
    return extract_nodes(soup, "lpage")


def collab(soup):
    return extract_nodes(soup, "collab")


def publisher_loc(soup):
    return extract_nodes(soup, "publisher-loc")


def publisher_name(soup):
    return extract_nodes(soup, "publisher-name")


def comment(soup):
    return extract_nodes(soup, "comment")


def element_citation(soup):
    return extract_nodes(soup, "element-citation")


def etal(soup):
    return extract_nodes(soup, "etal")


def pub_id(soup, pub_id_type=None):
    if pub_id_type:
        return extract_nodes(soup, "pub-id", attr="pub-id-type", value=pub_id_type)
    else:
        return extract_nodes(soup, "pub-id")


def source(soup):
    return extract_nodes(soup, "source")


def person_group(soup):
    return extract_nodes(soup, "person-group")


def surname(soup):
    return extract_nodes(soup, "surname")


def given_names(soup):
    return extract_nodes(soup, "given-names")


def suffix(soup):
    return extract_nodes(soup, "suffix")


def ext_link(soup, ext_link_type=None):
    if ext_link_type:
        return extract_nodes(
            soup, "ext-link", attr="ext-link-type", value=ext_link_type
        )
    else:
        return extract_nodes(soup, "ext-link")


def uri(soup):
    return extract_nodes(soup, "uri")


def edition(soup):
    return extract_nodes(soup, "edition")


def version(soup):
    return extract_nodes(soup, "version")


def chapter_title(soup):
    return extract_nodes(soup, "chapter-title")


def data_title(soup):
    return extract_nodes(soup, "data-title")


def conf_name(soup):
    return extract_nodes(soup, "conf-name")


def date_in_citation(soup):
    return extract_nodes(soup, "date-in-citation")


def patent(soup):
    return extract_nodes(soup, "patent")


#
# back
#


def back(soup):
    return first(extract_nodes(soup, "back"))


def app_group(soup):
    return extract_nodes(soup, "app-group")


def app(soup):
    return extract_nodes(soup, "app")


#
# body
#


def body(soup):
    return extract_nodes(soup, "body")


def article_body(soup):
    return first(extract_nodes(soup, "body"))


def sub_article(soup, article_type=None):
    return extract_nodes(soup, "sub-article", attr="article-type", value=article_type)


def editor_evaluation(soup):
    return first(sub_article(soup, "editor-report"))


def decision_letter(soup):
    tag = first(sub_article(soup, "article-commentary"))
    if not tag:
        tag = first(sub_article(soup, "decision-letter"))
    return tag


def author_response(soup):
    return first(sub_article(soup, "reply"))


def section(soup, sec_type=None):
    if sec_type:
        return extract_nodes(soup, "sec", attr="sec-type", value=sec_type)
    else:
        return extract_nodes(soup, "sec")


def paragraph(soup):
    return extract_nodes(soup, "p")


def table(soup):
    return extract_nodes(soup, "table")


def table_wrap_foot(soup):
    return extract_nodes(soup, "table-wrap-foot")


def disp_formula(soup):
    return extract_nodes(soup, "disp-formula")


def math(soup):
    return extract_nodes(soup, "math")


def boxed_text(soup):
    return extract_nodes(soup, "boxed-text")


def fig(soup):
    return extract_nodes(soup, "fig")


def fig_group(soup):
    return extract_nodes(soup, "fig-group")


def list(soup):  # Redefining `list` could be problematic
    return extract_nodes(soup, "list")


def list_item(soup):
    return extract_nodes(soup, "list-item")


#
# funding
#


def funding_group(soup):
    return extract_nodes(soup, "funding-group")


def award_group(soup):
    return extract_nodes(soup, "award-group")


def principal_award_recipient(soup):
    return extract_nodes(soup, "principal-award-recipient")


def string_name(soup):
    return extract_nodes(soup, "string-name")
